@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 """
 # Simple password manager
-# Copyright (c) 2011-2019 Michael Buesch <m@bues.ch>
+# Copyright (c) 2011-2021 Michael Buesch <m@bues.ch>
 # Licensed under the GNU/GPL version 2 or later.
 """
 
@@ -184,6 +184,8 @@ class PWMan(Cmd, metaclass=PWManMeta):
 		     commitClearsUndo=False, timeout=None):
 		super().__init__()
 
+		self.__isInteractive = False
+
 		if sys.flags.optimize >= 2:
 			# We need docstrings.
 			raise PWManError("pwman does not support "
@@ -230,9 +232,8 @@ class PWMan(Cmd, metaclass=PWManMeta):
 		self._timeout.poke()
 
 	def default(self, line):
-		self.__err(None,
-			   "Unknown command: %s\n"
-			   "Type 'help' for more help." % line)
+		extra = "\nType 'help' for more help." if self.__isInteractive else ""
+		self.__err(None, "Unknown command: %s%s" % (line, extra))
 
 	def emptyline(self):
 		self._timeout.poke()
@@ -251,6 +252,23 @@ class PWMan(Cmd, metaclass=PWManMeta):
 							       text)
 		return []
 
+	@completion
+	def __complete_category_title_item(self, text, line, begidx, endidx):
+		paramIdx = self._calcParamIndex(line, endidx)
+		if paramIdx in (0, 1):
+			return self.__complete_category_title(text, line, begidx, endidx)
+		category, title, item = self._getComplParams(line, 0, 3)
+		cmpl = []
+		if paramIdx == 2:
+			cmpl.extend(escapeCmd(n) + " "
+				    for n in ("user", "password", "bulk", "totpkey")
+				    if n.startswith(item.lower()))
+		cmpl.extend(self.__getEntryAttrCompletions(category, title, item,
+							   doName=(paramIdx == 2),
+							   doData=False,
+							   text=text))
+		return cmpl
+
 	def __getCategoryCompletions(self, text):
 		return [ escapeCmd(n) + " "
 			 for n in self.__db.getCategoryNames()
@@ -261,6 +279,22 @@ class PWMan(Cmd, metaclass=PWManMeta):
 			 for t in self.__db.getEntryTitles(category)
 			 if t.lower().startswith(text.lower()) ]
 
+	def __getEntryAttrCompletions(self, category, title, name, doName, doData, text):
+		if category and title:
+			entry = self.__db.getEntry(category, title)
+			if entry:
+				if doName: # complete name
+					entryAttrs = self.__db.getEntryAttrs(entry)
+					if entryAttrs:
+						return [ escapeCmd(entryAttr.name) + " "
+							 for entryAttr in entryAttrs
+							 if entryAttr.name.lower().startswith(name.lower()) ]
+				elif doData: # complete data
+					entryAttr = self.__db.getEntryAttr(entry, name)
+					if entryAttr:
+						return [ escapeCmd(entryAttr.data) + " " ]
+		return []
+
 	def __getPathCompletions(self, text):
 		"""Return an escaped file system path completion.
 		'text' is the unescaped partial path string.
@@ -293,7 +327,6 @@ class PWMan(Cmd, metaclass=PWManMeta):
 		("list", ("ls", "cat"), "List/print entry contents"),
 		("find", ("f",), "Search the database for patterns"),
 		("totp", ("t",), "Generate TOTP token"),
-		("totp_key", ("tk",), "Show TOTP key and parameters"),
 		("diff", (), "Show the database differences"),
 	)
 
@@ -403,30 +436,67 @@ class PWMan(Cmd, metaclass=PWManMeta):
 
 	def do_list(self, params):
 		"""--- Print a listing ---
-		Command: list [category] [title]\n
+		Command: list [category] [title] [item]\n
 		If a category is given as parameter, list the 
 		contents of the category. If category and entry
-		are given, list the contents of the entry.\n
+		are given, list the contents of the entry.
+		If item is given, then only list one specific content item.
+		Item may be one of: user, password, bulk, totpkey or any attribute name.\n
 		Aliases: ls cat"""
-		category, title = self._getParams(params, 0, 2)
-		if not category and not title:
+		category, title, item = self._getParams(params, 0, 3)
+		if not category and not title and not item:
 			self.__info(None, "Categories:")
 			self.__info(None, "\t" + "\n\t".join(self.__db.getCategoryNames()))
-		elif category and not title:
+		elif category and not title and not item:
 			self.__info(None, "Entries in category '%s':" % category)
 			self.__info(None, "\t" + "\n\t".join(self.__db.getEntryTitles(category)))
-		elif category and title:
+		elif category and title and not item:
 			entry = self.__db.getEntry(category, title)
 			if entry:
 				self.__info(None, self.__db.dumpEntry(entry))
 			else:
 				self.__err("list", "'%s/%s' not found" % (category, title))
+		elif category and title and item:
+			entry = self.__db.getEntry(category, title)
+			if entry:
+				if item == "user":
+					if not entry.user:
+						self.__err("list", "'%s/%s' has no 'user' field." % (
+							   category, title))
+					self.__info(None, entry.user)
+				elif item == "password":
+					if not entry.pw:
+						self.__err("list", "'%s/%s' has no 'password' field." % (
+							   category, title))
+					self.__info(None, entry.pw)
+				elif item == "bulk":
+					bulk = self.__db.getEntryBulk(entry)
+					if not bulk:
+						self.__err("list", "'%s/%s' has no 'bulk' field." % (
+							   category, title))
+					self.__info(None, bulk.data)
+				elif item == "totpkey":
+					entryTotp = self.__db.getEntryTotp(entry)
+					if not entryTotp:
+						self.__err("list", "'%s/%s' has no 'TOTP key'." % (
+							   category, title))
+					self.__info(None, "TOTP key:     %s (base32 encoding)" % entryTotp.key)
+					self.__info(None, "TOTP digits:  %d" % entryTotp.digits)
+					self.__info(None, "TOTP hash:    %s" % entryTotp.hmacHash)
+				else: # attribute
+					attr = self.__db.getEntryAttr(entry, item)
+					if not attr:
+						self.__err("list", "'%s/%s' has no attribute '%s'." % (
+							   category, title, item))
+					self.__info(None, attr.data)
+			else:
+				self.__err("list", "'%s/%s' not found" % (category, title))
 		else:
 			self.__err("list", "Invalid parameter")
 	do_ls = do_list
 	do_cat = do_list
 
-	complete_list = __complete_category_title
+	complete_list = __complete_category_title_item
 	complete_ls = complete_list
 	complete_cat = complete_list
 
@@ -905,34 +975,6 @@ class PWMan(Cmd, metaclass=PWManMeta):
 	complete_totp = __complete_category_title
 	complete_t = complete_totp
 
-	def do_totp_key(self, params):
-		"""--- Show TOTP key and parameters ---
-		Command: totp_key category title\n
-		Show Time-Based One-Time Password Algorithm key and parameters.\n
-		Aliases: tk"""
-		category, title = self._getParams(params, 0, 2)
-		if not category:
-			self.__err("totp_key", "Category parameter is required.")
-		if not title:
-			self.__err("totp_key", "Title parameter is required.")
-		entry = self.__db.getEntry(category, title)
-		if not entry:
-			self.__err("totp_key", "'%s/%s' not found" % (category, title))
-		entryTotp = self.__db.getEntryTotp(entry)
-		enc = "  (base32 encoding)"
-		if not entryTotp:
-			entryTotp = PWManEntryTOTP(key="--- none ---",
-						   digits=6,
-						   hmacHash="SHA1")
-			enc = ""
-		self.__info(None, "TOTP key:     %s%s" % (entryTotp.key, enc))
-		self.__info(None, "TOTP digits:  %d" % entryTotp.digits)
-		self.__info(None, "TOTP hash:    %s" % entryTotp.hmacHash)
-	do_tk = do_totp_key
-
-	complete_totp_key = __complete_category_title
-	complete_tk = complete_totp_key
-
 	def do_edit_totp(self, params):
 		"""--- Edit TOTP key and parameters ---
 		Command: edit_totp category title [KEY] [DIGITS] [HASH]\n
@@ -1024,20 +1066,10 @@ class PWMan(Cmd, metaclass=PWManMeta):
 		if paramIdx in (0, 1):
 			return self.__complete_category_title(text, line, begidx, endidx)
 		category, title, name = self._getComplParams(line, 0, 3)
-		if category and title:
-			entry = self.__db.getEntry(category, title)
-			if entry:
-				if paramIdx == 2: # name
-					entryAttrs = self.__db.getEntryAttrs(entry)
-					if entryAttrs:
-						return [ escapeCmd(entryAttr.name) + " "
-							 for entryAttr in entryAttrs
-							 if entryAttr.name.lower().startswith(name.lower()) ]
-				elif paramIdx == 3: # data
-					entryAttr = self.__db.getEntryAttr(entry, name)
-					if entryAttr:
-						return [ escapeCmd(entryAttr.data) + " " ]
-		return []
+		return self.__getEntryAttrCompletions(category, title, name,
+						      doName=(paramIdx == 2),
+						      doData=(paramIdx == 3),
+						      text=text)
 	complete_ea = complete_edit_attr
 
 	__diff_opts = ("-u", "-c", "-n")
@@ -1329,29 +1361,36 @@ class PWMan(Cmd, metaclass=PWManMeta):
 		self.__db.flunkDirty()
 
 	def interactive(self):
-		while True:
-			try:
-				self.cmdloop()
-				break
-			except (self.Quit) as e:
-				if self.__mayQuit():
-					self.do_cls("")
+		self.__isInteractive = True
+		try:
+			while True:
+				try:
+					self.cmdloop()
 					break
-			except EscapeError as e:
-				self.__warn(None, str(e))
-			except self.CommandError as e:
-				print(str(e), file=sys.stderr)
-			except (KeyboardInterrupt, EOFError) as e:
-				print("")
-			except (CSQLError) as e:
-				self.__warn(None, "SQL error: %s" % str(e))
+				except self.Quit as e:
+					if self.__mayQuit():
+						self.do_cls("")
+						break
+				except EscapeError as e:
+					self.__warn(None, str(e))
+				except self.CommandError as e:
+					print(str(e), file=sys.stderr)
+				except (KeyboardInterrupt, EOFError) as e:
+					print("")
+				except CSQLError as e:
+					self.__warn(None, "SQL error: %s" % str(e))
+		finally:
+			self.__isInteractive = False
 
 	def runOneCommand(self, command):
+		self.__isInteractive = False
 		try:
 			self.onecmd(command)
+		except self.Quit as e:
+			raise PWManError("Quit command executed in non-interactive mode.")
 		except (EscapeError, self.CommandError) as e:
 			raise PWManError(str(e))
 		except (KeyboardInterrupt, EOFError) as e:
-			raise PWManError("Interrupted")
-		except (CSQLError) as e:
+			raise PWManError("Interrupted.")
+		except CSQLError as e:
 			raise PWManError("SQL error: %s" % str(e))
