@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 """
 # Simple password manager
-# Copyright (c) 2011-2019 Michael Buesch <m@bues.ch>
+# Copyright (c) 2011-2021 Michael Buesch <m@bues.ch>
 # Licensed under the GNU/GPL version 2 or later.
 """
 
@@ -11,6 +11,7 @@ import os
 import sys
 
 __all__ = [
+	"str2bool",
 	"osIsWindows",
 	"osIsPosix",
 	"uniq",
@@ -22,6 +23,19 @@ __all__ = [
 osIsWindows = os.name == "nt" or os.name == "ce"
 osIsPosix = os.name == "posix"
 
+def str2bool(string, default=False):
+	s = string.lower().strip()
+	if not s:
+		return default
+	if s in ("true", "yes", "on", "1"):
+		return True
+	if s in ("false", "no", "off", "0"):
+		return False
+	try:
+		return bool(int(s))
+	except ValueError:
+		return default
+
 def uniq(l, sort=True):
 	l = list(set(l))
 	if sort:
@@ -48,17 +62,23 @@ def clearScreen():
 		curses.endwin()
 	stdout("\x1B[2J\x1B[0;0f")
 
+def _do_getpass(prompt):
+	if str2bool(os.getenv("PWMAN_RAWGETPASS", "")):
+		return input(prompt)
+	else:
+		return getpass.getpass(prompt)
+
 def readPassphrase(prompt, verify=False):
 	if verify:
 		prompt = "[New] " + prompt
 	try:
 		while True:
-			p0 = getpass.getpass(prompt + ": ")
+			p0 = _do_getpass(prompt + ": ")
 			if not p0:
 				continue
 			if not verify:
 				return p0
-			p1 = getpass.getpass(prompt + " (verify): ")
+			p1 = _do_getpass(prompt + " (verify): ")
 			if p0 == p1:
 				return p0
 			print("Passwords don't match. Try again...",
